import aiobotocore
import asyncio
import os
import struct
import sys


def _unitize(value):
    unit = ["", "K", "M", "G", "T"]
    while value > 100 and unit:
        value /= 1024
        unit.pop(0)
    return value, unit[0]


def _speed(sub):
    value = (sub[-1][0] - sub[0][0]) / (sub[-1][1] - sub[0][1])
    value, unit = _unitize(value)
    return f"{value:.1f}{unit}B/s"


class S3Log:
    def __init__(
        self,
        bucket,
        prefix,
        script,
        region,
        access_key_id,
        secret_access_key,
        concurrency,
        progress,
    ):
        self._bucket = bucket
        self._prefix = prefix
        self._script = script
        self._aws_region = region
        self._aws_access_key_id = access_key_id
        self._aws_secret_access_key = secret_access_key
        self._concurrency = concurrency
        self._show_progress = progress

        self._loop = asyncio.get_event_loop()
        self._total_lines = 0
        self._total_received = 0
        self._total_processed = 0
        self._size_queue = []
        self._q = asyncio.Queue()
        self._tasks_queue = asyncio.Queue()

        print(
            f"Processing logs from {self._bucket}/{self._prefix} in {self._aws_region}",
            file=sys.stderr,
        )
        print(f"Concurrency: {self._concurrency}", file=sys.stderr)
        print(f"Show progress: {self._show_progress}", file=sys.stderr)

    async def _wait(self):
        while True:
            task = await self._tasks_queue.get()
            if not task:
                break
            await task

    async def feed(self, client, key, proc):
        response = await client.get_object(Bucket=self._bucket, Key=key)
        # this will ensure the connection is correctly re-used/closed
        async with response["Body"] as stream:
            size = 0
            while True:
                chunk = await stream.readany()
                if not chunk:
                    break
                proc.stdin.write(chunk)
                len_chunk = len(chunk)
                size += len_chunk
                self._total_received += len_chunk
                if size > 65536:
                    size = 0
                    await proc.stdin.drain()
            proc.stdin.write(b"\x00")
            await proc.stdin.drain()
            result = await proc.stdout.read(16)
            await self._q.put(proc)
            lines, size = struct.unpack("QQ", result)
            self._total_lines += lines
            self._total_processed += size

    async def _status(self):
        start = [(0, self._loop.time())]
        self._size_queue.extend(start)
        while True:
            await asyncio.sleep(1)
            self._size_queue.append((self._total_received, self._loop.time()))
            if len(self._size_queue) > 60:
                self._size_queue.pop(0)
            print(
                f"MA5: {_speed(self._size_queue[-5:])}\t",
                f"MA20: {_speed(self._size_queue[-20:])}\t",
                f"MA60: {_speed(self._size_queue)}\t",
                f"AVG: {_speed(start + self._size_queue[-1:])}\t",
                f"JSON: {self._total_processed / self._total_lines if self._total_lines else 0:,.0f}B",
                "Size: {:.1f}{}B".format(*_unitize(self._total_processed)),
                file=sys.stderr,
            )

    async def _run(self):
        for i in range(self._concurrency):
            proc = await asyncio.create_subprocess_exec(
                sys.executable,
                os.path.join(
                    os.path.dirname(os.path.abspath(__file__)), "s3log_worker.py"
                ),
                self._script,
                stdin=asyncio.subprocess.PIPE,
                stdout=asyncio.subprocess.PIPE,
                stderr=sys.stdout,
            )
            await self._q.put(proc)

        session = aiobotocore.get_session(loop=self._loop)
        async with session.create_client(
            "s3",
            region_name=self._aws_region,
            aws_secret_access_key=self._aws_secret_access_key,
            aws_access_key_id=self._aws_access_key_id,
        ) as client:
            if self._show_progress:
                self._loop.create_task(self._status())
            waiter = self._loop.create_task(self._wait())
            # # list s3 objects using paginator
            paginator = client.get_paginator("list_objects")
            async for result in paginator.paginate(
                Bucket=self._bucket, Prefix=self._prefix
            ):
                for c in result.get("Contents", []):
                    key = c["Key"]
                    print(f"Processing key: {key}", file=sys.stderr)
                    proc = await self._q.get()
                    await self._tasks_queue.put(
                        self._loop.create_task(self.feed(client, key, proc))
                    )
            await self._tasks_queue.put(None)
            await waiter

    def run(self):
        self._loop.run_until_complete(self._run())



from collections import defaultdict
from packaging import version
import yaml
import re

from gen3utils.manifest.manifest_validator import get_manifest_version
from gen3utils.etl.dd_utils import init_dictionary
from gen3utils.errors import MappingError, PropertiesError, PathError, FieldError


class Prop:
    def __init__(self, name):
        self.name = name


class Index:
    def __init__(self, name, underscore=False):
        self.name = name
        if underscore:
            id_name = "_{}_id".format(name)
        else:
            id_name = "{}_id".format(name)
        self.props = {id_name: Prop(id_name)}


def validate_joining_list_props(props_list, recorded_errors, existing_indices):
    if type(props_list) is list:
        for prop in props_list:
            if "index" in prop and "join_on" in prop:
                for real_prop in prop.get("props"):
                    validate_joining_prop(
                        real_prop,
                        recorded_errors,
                        existing_indices.get(prop.get("index")),
                    )


def validate_list_props(
    props_list,
    labels_to_back_refs,
    nodes_with_props,
    recorded_errors,
    grouping_path,
    index,
    nodes_for_category=None,
):
    if not nodes_for_category:
        nodes_for_category = []

    if type(props_list) is list:
        for prop in props_list:
            if "path" in prop and "props" in prop:  # flatten_props
                for real_prop in prop.get("props"):
                    new_props = validate_prop(
                        real_prop,
                        labels_to_back_refs,
                        nodes_with_props,
                        recorded_errors,
                        prop.get("path", grouping_path),
                    )
                    index.props.update({p.name: p for p in new_props})
            elif "index" in prop and "join_on" in prop:  # joining_props
                # joining_props does not require path (considering it later after having all indices)
                return
            else:
                new_props = validate_prop(
                    prop,
                    labels_to_back_refs,
                    nodes_with_props,
                    recorded_errors,
                    grouping_path,
                    nodes_for_category,
                )
                index.props.update({p.name: p for p in new_props})
            # joining_props which contain join_on and index will be validated after all indices are walked through
    elif type(props_list) is dict:
        for k, v in props_list.items():
            if k in labels_to_back_refs.keys():
                for prop in v.get("props"):
                    new_props = validate_prop(
                        prop,
                        labels_to_back_refs,
                        nodes_with_props,
                        recorded_errors,
                        labels_to_back_refs.get(k),
                    )
                    index.props.update({p.name: p for p in new_props})


def validate_joining_prop(json_obj, recorded_errors, joining_index):
    name = validate_name(json_obj, recorded_errors)
    validate_fn(json_obj, recorded_errors)
    validate_joining_src(json_obj, recorded_errors, joining_index.props)
    return Prop(name)


def validate_prop(
    json_obj,
    labels_to_back_refs,
    nodes_with_props,
    recorded_errors,
    grouping_path=None,
    nodes_for_category=None,
):
    if not nodes_for_category:
        nodes_for_category = []

    names = validate_path(
        json_obj,
        grouping_path,
        recorded_errors,
        labels_to_back_refs,
        nodes_with_props,
        nodes_for_category,
    )
    if len(names) == 0:
        names.append(
            validate_name_src(
                json_obj,
                json_obj.get("path", grouping_path),
                recorded_errors,
                nodes_with_props,
                labels_to_back_refs,
                nodes_for_category,
            )
        )

    props = [Prop(n) for n in names]
    return props


def validate_joining_src(json_obj, recorded_errors, joining_props):
    src = json_obj.get("src", json_obj.get("name"))
    alternate_src = json_obj.get("alternate_src", json_obj.get("name"))
    if src is not None:
        if src not in joining_props:
            if alternate_src in joining_props:
                print("WARN: Utilizing alternate src")
                return
            recorded_errors.append(
                FieldError(
                    'src field "{}" (declared in "{}") is not found in joining index.'.format(
                        src, json_obj
                    )
                )
            )
    else:
        recorded_errors.append(
            FieldError('Missing source field for "{}"'.format(json_obj))
        )


def validate_fn(json_obj, recorded_errors):
    fn = json_obj.get("fn")
    if fn is not None:
        if fn not in ["set", "count", "list", "sum", "min", "max"]:
            recorded_errors.append(
                MappingError(
                    '"{}" function (declared in "{}") is not supported in ETL'.format(
                        fn, json_obj
                    ),
                    "Function",
                )
            )
    return fn


def validate_name(json_obj, recorded_errors):
    name = json_obj.get("name")
    if name is None or name == "":
        recorded_errors.append(
            PropertiesError(
                'Name is missing or empty string for mapping property "{}".'.format(
                    json_obj
                )
            )
        )
    return name


def validate_name_src(
    json_obj,
    path,
    recorded_errors,
    nodes_with_props,
    labels_to_back_refs=None,
    nodes_for_category=None,
):
    if not labels_to_back_refs:
        labels_to_back_refs = {}
    if not nodes_for_category:
        nodes_for_category = []

    name = validate_name(json_obj, recorded_errors)
    fn = validate_fn(json_obj, recorded_errors)
    src = json_obj.get("src", name)
    if not src:
        return name
    if not path and not nodes_for_category:
        recorded_errors.append(
            FieldError(
                'src field must be specified with a path for "{}"'.format(json_obj)
            )
        )
    else:
        valid_fields = {"source_node"}  # built-in fields
        if path:
            path_items = path.split(".")
            valid_fields.update(nodes_with_props.get(path_items[-1], []))
        for node_name in nodes_for_category:
            node_backref = labels_to_back_refs[node_name]
            node_props = nodes_with_props.get(node_backref, [])
            valid_fields.update(node_props)
        if fn != "count" and src not in valid_fields:
            recorded_errors.append(
                FieldError(
                    'src field "{}" (declared in {} "{}") is not found in given dictionary.'.format(
                        src, f'"{path}"' if path else "a collector index", json_obj
                    )
                )
            )
    return name


def validate_path(
    json_obj,
    grouping_path,
    recorded_errors,
    labels_to_back_refs,
    nodes_with_props,
    nodes_for_category=None,
):
    if not nodes_for_category:
        nodes_for_category = []
    nodes_for_category_backrefs = [labels_to_back_refs[n] for n in nodes_for_category]

    path = json_obj.get("path", grouping_path)
    names = []
    if path:
        nodes_for_category_backrefs.append(path)
    if not nodes_for_category_backrefs:
        recorded_errors.append(
            PropertiesError(
                'Missing path declaration for the property "{}".'.format(json_obj)
            )
        )
    else:
        for path in nodes_for_category_backrefs:
            # handle format "node1[id].node2[id]":
            path_items = path.split(".")
            if "_ANY" in path_items:
                path_items.remove("_ANY")
            for item in path_items:
                # get the edge name and the property definition from line:
                # subjects[subject_id:id,project_id]
                [edge, str_fields] = (
                    list(filter(None, re.split(r"[\[\]]", item)))
                    if "[" in item
                    else [item, None]
                )
                if edge not in labels_to_back_refs.values():
                    recorded_errors.append(PathError(path))
                if str_fields is not None:
                    fields = str_fields.split(",")
                    for f in fields:
                        name, src = (
                            [p.strip() for p in f.split(":")]
                            if ":" in f
                            else [f.strip()] * 2
                        )
                        names.append(
                            validate_name_src(
                                {"name": name, "src": src},
                                edge,
                                recorded_errors,
                                nodes_with_props,
                            )
                        )
    return names


def get_all_nodes(model):
    labels_to_back_refs = {}
    """
    dictionary from label to back_ref
    {
        "subject": "subjects"
    }
    """
    nodes_with_props = {}
    """
    dictionary nodes (with plural) to all properties
    {
        "subjects": [
            "submitter_id",
            "project_id",
            "species"
        ]
    }
    """
    categories_to_labels = defaultdict(list)
    """
    group label by its category
    {
        "data_file": [
            submitted_aligned_reads,
            submitted_unaligned_reads
        ]
    }
    """
    all_classes = model.Node.get_subclasses()
    for n in all_classes:
        present_links = n._pg_edges
        present_props = n.__pg_properties__
        present_node_props = list(present_props.keys()) + ["id"]
        category = n._dictionary.get("category")
        categories_to_labels[category].append(n.label)
        backref = ""
        if len(present_links) > 0:
            backref = list(present_links.values())[0].get("backref")
            labels_to_back_refs[n.label] = backref
        nodes_with_props.update({backref: present_node_props})
    return labels_to_back_refs, nodes_with_props, categories_to_labels


def check_mapping_format(mappings, recorded_errors):
    # TODO add more checks to this
    if "mappings" not in mappings:
        recorded_errors.append(
            MappingError('etlMapping file does not contain "mappings"', "format")
        )
        return recorded_errors
    for m in mappings.get("mappings"):
        if "doc_type" not in m:
            recorded_errors.append(
                MappingError(
                    'Mapping "{}" does not contain "doc_type"'.format(m.get("name")),
                    "format",
                )
            )
    return recorded_errors


def check_mapping_constraints(mappings, model, recorded_errors, underscore):
    labels_to_back_refs, nodes_with_props, categories_to_labels = get_all_nodes(model)
    indices = {}
    for m in mappings.get("mappings"):
        index = Index(m.get("doc_type"), underscore)
        indices[index.name] = index
        category = m.get("category")
        node_name = m.get("root")
        for key, value in m.items():
            if key.endswith("props"):
                root_path = (
                    labels_to_back_refs.get(node_name) if key == "props" else None
                )
                nodes_for_category = categories_to_labels.get(category, [])
                validate_list_props(
                    value,
                    labels_to_back_refs,
                    nodes_with_props,
                    recorded_errors,
                    root_path,
                    index,
                    nodes_for_category,
                )
        if (
            m.get("type") == "aggregator"
            and m.get("root") != "project"
            and "project_id" not in index.props
        ):
            recorded_errors.append(
                PropertiesError(f"project_id is not found in index {index.name}")
            )

    for m in mappings.get("mappings"):
        joining_props = m.get("joining_props", [])
        validate_joining_list_props(joining_props, recorded_errors, indices)
    return recorded_errors


def is_release_tag(parsed_version):
    """
    Args:
        parsed_version: releases.version
    Returns:
        bool: True if version's major number is >= 2019. This is true of
        our current monthly release versions (e.g. 2020.05, 2019.11) but not
        true of our standard semver versions (e.g. 2.33.0)
    """
    return parsed_version >= version.parse("2019.0")


def validate_mapping(dictionary_url, mapping_file, manifest):
    dictionary, model = init_dictionary(dictionary_url)
    with open(mapping_file) as f:
        mappings = yaml.safe_load(f)

    # If using tube >= 0.4.0 or >= 2020.10, {doc_type}_id fields have a prefixed underscore.
    # https://github.com/uc-cdis/tube/releases/tag/0.4.0
    tube_version = get_manifest_version(
        manifest["versions"], "tube", release_tag_are_branches=False, warn=False
    )
    underscore = False
    if tube_version is not None and type(tube_version) != str:  # str if branch
        if not is_release_tag(tube_version) and tube_version >= version.parse("0.4.0"):
            underscore = True
        elif tube_version >= version.parse("2020.10"):
            underscore = True

    recorded_errors = check_mapping_format(mappings, [])
    if len(recorded_errors) > 0:
        return recorded_errors

    return check_mapping_constraints(mappings, model, recorded_errors, underscore)



import re
import yaml
import json
import os

from cdislogging import get_logger

from gen3utils.assertion import assert_and_log
from gen3utils.etl.dd_utils import init_dictionary
from gen3utils.errors import FieldSyntaxError, FieldError
from gen3utils.deployment_changes.generate_comment import submit_comment


logger = get_logger("validate-portal-config", log_level="info")


def val_gitops(data_dictionary, etl_mapping, gitops):
    with open(gitops, "r") as f:
        gitops_config = json.loads(f.read())
    ok = validate_gitops_syntax(gitops_config)
    if not ok:
        raise AssertionError(
            "Portal configuration failed. See errors in previous logs."
        )

    ok = validate_against_dictionary(gitops_config, data_dictionary)
    recorded_errors = validate_against_etl(gitops_config, etl_mapping)
    return recorded_errors, ok


def validate_gitops_syntax(gitops):
    """
    Validates the syntax of gitops.json by checking for required fields
    Args:
        gitops (dict): gitops.json config

    Returns:
        Error: Returns the error if it exists, else returns None.

    TODO
    Currently only checks syntax needed for etl mapping and dictionary validation.
    """
    graphql = gitops.get("graphql")
    ok = assert_and_log(graphql, FieldSyntaxError("graphql"))

    if graphql:
        ok = ok and assert_and_log(
            "boardCounts" in graphql, FieldSyntaxError("graphql.boardCounts")
        )
        boardcounts = graphql.get("boardCounts", [])
        if boardcounts:
            for item in boardcounts:
                checks = ["graphql", "name", "plural"]
                ok = check_required_fields("graphql.boardCounts", checks, item, ok)

        ok = ok and assert_and_log(
            "chartCounts" in graphql, FieldSyntaxError("graphql.chartCounts")
        )

    components = gitops.get("components")
    ok = ok and assert_and_log(components, FieldSyntaxError("components"))
    if components:
        index = components.get("index")
        ok = ok and assert_and_log(index, FieldSyntaxError("components.index"))
        if index:
            homepage = index.get("homepageChartNodes", [])
            ok = ok and assert_and_log(
                index, FieldSyntaxError("components.homepageChartNodes")
            )
            for item in homepage:
                checks = ["node", "name"]
                ok = check_required_fields(
                    "components.index.homepage", checks, item, ok
                )

        # footerLogos is optional, but when present, it must be an array (list)
        footerLogos = components.get("footerLogos", [])
        ok = ok and assert_and_log(
            isinstance(footerLogos, list),
            FieldError("footerLogos must be an array if presents in the config"),
        )

    explorer_enabled = gitops.get("featureFlags", {}).get("explorer", True)
    explorerconfig = gitops.get("explorerConfig")
    configs = []
    if not explorerconfig:
        dataConfig = gitops.get("dataExplorerConfig")
        fileConfig = gitops.get("fileExplorerConfig")
        ok = ok and assert_and_log(
            dataConfig or not explorer_enabled, FieldSyntaxError("(data)explorerConfig")
        )
        if dataConfig:
            configs.append(dataConfig)
        if fileConfig:
            configs.append(fileConfig)
    else:
        configs = explorerconfig

    for exp_config in configs:
        filters = exp_config.get("filters")
        ok = ok and assert_and_log(filters, FieldSyntaxError("explorerConfig.filters"))
        if filters:
            tabs = filters.get("tabs", [])
            ok = ok and assert_and_log(
                tabs, FieldSyntaxError("explorerConfig.filters.tabs")
            )
            for tab in tabs:
                checks = ["title", "fields"]
                ok = check_required_fields(
                    "explorerConfig.filters.tab", checks, tab, ok
                )

        guppy = exp_config.get("guppyConfig")
        ok = ok and assert_and_log(
            guppy, FieldSyntaxError("explorerConfig.guppyConfig")
        )
        buttons = exp_config.get("buttons", [])
        manifest_mapping = None
        if guppy:
            ok = ok and assert_and_log(
                guppy.get("dataType"),
                FieldSyntaxError("explorerConfig.guppyConfig.dataType"),
            )
            manifest_mapping = guppy.get("manifestMapping")

        val_mapping = False
        for button in buttons:
            if button.get("enabled") and button.get("type") == "manifest":
                val_mapping = True
        if manifest_mapping and val_mapping:
            checks = [
                "resourceIndexType",
                "resourceIdField",
                "referenceIdFieldInResourceIndex",
                "referenceIdFieldInDataIndex",
            ]
            ok = check_required_fields(
                "explorerConfig.guppyConfig.manifestMapping",
                checks,
                manifest_mapping,
                ok,
            )

    study_viewer = gitops.get("studyViewerConfig")

    if study_viewer:
        checks = [
            "dataType",
            "listItemConfig",
            "rowAccessor",
        ]
        for viewer in study_viewer:
            ok = check_required_fields("studyViewerConfig", checks, viewer, ok)

    return ok


def check_required_fields(path, checks, field, ok):
    for check in checks:
        ok = ok and assert_and_log(
            field.get(check), FieldSyntaxError(f"{path}.{check}")
        )
    return ok


def check_field_value(path, checks, accepted_values, errors):
    for check in checks:
        if check not in accepted_values:
            errors.append(
                FieldError(
                    "Field [{}] in {} not found in etlMapping".format(check, path)
                )
            )


def _validate_itemConfig(item_config, props, typ, errors):
    check_field_value(
        "studyViewerConfig.blockFields.{}ItemConfig".format(typ),
        item_config.get("blockFields", []),
        props,
        errors,
    )
    check_field_value(
        "studyViewerConfig.tableFields.{}ItemConfig".format(typ),
        item_config.get("tableFields", []),
        props,
        errors,
    )


def _validate_studyViewer_datatypes(
    viewer, datatype, type_prop_map, row_accessor, errors
):
    dtype = viewer.get(datatype)
    props = type_prop_map.get(dtype, [])
    if not props:
        errors.append(
            FieldError(
                "Field [{}] in studyViewerConfig.{} not found in etlMapping".format(
                    dtype, datatype
                )
            )
        )
    if row_accessor not in props:
        errors.append(
            FieldError(
                "rowAccessor [{}] not found in index with type {}".format(
                    row_accessor, dtype
                )
            )
        )


def _validate_studyViewerConfig_helper(viewer, type_prop_map, errors):
    dtype = viewer.get("dataType")
    checks = ["dataType"]
    for optional in ["fileDataType", "docDataType"]:
        if viewer.get(optional):
            checks.append(optional)

    for datatype in checks:
        _validate_studyViewer_datatypes(
            viewer, datatype, type_prop_map, viewer["rowAccessor"], errors
        )
    props = type_prop_map.get(dtype, [])
    listitemconfig = viewer.get("listItemConfig")
    _validate_itemConfig(listitemconfig, props, "list", errors)
    if viewer.get("singleItemConfig"):
        _validate_itemConfig(viewer.get("singleItemConfig"), props, "single", errors)


def validate_studyViewerConfig(studyviewer, type_prop_map, errors):
    for viewer in studyviewer:
        _validate_studyViewerConfig_helper(viewer, type_prop_map, errors)


def _validate_explorerConfig_helper(explorer_config, type_prop_map, errors):

    datatype = explorer_config["guppyConfig"]["dataType"]
    props = type_prop_map.get(datatype, [])
    if not props:
        errors.append(
            FieldError(
                "Field [{}] in explorerConfig.guppyConfig.dataType not found in etlMapping".format(
                    datatype
                )
            )
        )

    tabs = explorer_config["filters"]["tabs"]
    for tab in tabs:
        check_field_value(
            "explorerConfig.filters.tabs.fields", tab.get("fields", []), props, errors
        )

    table = explorer_config["table"]
    if table["enabled"]:
        check_field_value(
            "explorerConfig.table.fields", table.get("fields", []), props, errors
        )

    manifest_map = explorer_config.get("manifestMapping")
    if manifest_map and manifest_map.get("resourceIndexType"):
        resource_props = type_prop_map.get(manifest_map.get("resourceIndexType"))
        if not resource_props:
            errors.append(
                FieldError(
                    "Field [{}] in manifestMapping.resourceIndexType not found in etlMapping".format(
                        manifest_map.get("resourceIndexType")
                    )
                )
            )
        elif manifest_map.get("resourceIdField") not in resource_props:
            errors.append(
                FieldError(
                    "Field [{}] in manifestMapping.resourceIdField not found in etlMapping".format(
                        manifest_map.get("resourceIdField")
                    )
                )
            )
        # TODO
        # Also consider fields referenceIdFieldInResourceIndex and referenceIdFieldInDataIndex


def validate_explorerConfig(gitops, type_prop_map, errors):
    explorer_configs = gitops.get("explorerConfig", [])
    # if explorerConfig exists, ignores (data/files)explorerConfig
    if not explorer_configs:
        if "dataExplorerConfig" in gitops:
            explorer_configs.append(gitops["dataExplorerConfig"])
        if "fileExplorerConfig" in gitops:
            explorer_configs.append(gitops["fileExplorerConfig"])
    for config in explorer_configs:
        _validate_explorerConfig_helper(config, type_prop_map, errors)
    return errors


def validate_against_etl(gitops, mapping_file):
    """
    Validates gitops.json configuration against an etlMapping
    Args:
        gitops (dict): gitops.json config
        mapping_file (str): path to etlMapping file

    Returns:
        Error: Returns a list of any errors encountered.

    """
    with open(mapping_file) as f:
        mappings = yaml.safe_load(f)
    mapping = mappings.get("mappings")
    type_prop_map = map_all_ES_index_props(mapping)
    errors = validate_explorerConfig(gitops, type_prop_map, [])
    studyviewer = gitops.get("studyViewerConfig")
    if studyviewer:
        validate_studyViewerConfig(studyviewer, type_prop_map, errors)

    return errors


def map_all_ES_index_props(mapping):
    """
    Args:
        mapping (dict): The etlMapping to parse

    returns:
        A mapping between each index type and all it's properties
    """
    all_prop_map = {}
    for index in mapping:
        index_props = []
        props = index.get("props")
        index_props.extend(_extract_props(props))
        agg_props = index.get("aggregated_props")
        index_props.extend(_extract_props(agg_props))
        join_props = index.get("joining_props", [])
        for indx in join_props:
            index_props.extend(_extract_props(index.get("props")))
        inject_props = index.get("injecting_props", {})
        for node, props in inject_props.items():
            index_props.extend(_extract_props(props.get("props")))
        flat_props = index.get("flatten_props", [])
        for node in flat_props:
            index_props.extend(_extract_props(node.get("props")))
        parent_props = index.get("parent_props")
        if parent_props:
            for prop in parent_props:
                # handle format "node1[id].node2[id]":
                path_items = prop.get("path").split(".")
                if "_ANY" in path_items:
                    path_items.remove("_ANY")
                # get the edge name and the property definition from line:
                # subjects[subject_id:id,project_id]
                for item in path_items:
                    [_, str_props] = (
                        list(filter(None, re.split(r"[\[\]]", item)))
                        if "[" in item
                        else [item, None]
                    )
                    if str_props is not None:
                        props = str_props.split(",")
                        index_props.extend(
                            [
                                p.split(":")[0].strip()
                                if p.find(":") != -1
                                else p.strip()
                                for p in props
                            ]
                        )

        all_prop_map[index.get("doc_type")] = set(index_props)

    return all_prop_map


def _extract_props(props_to_extract):
    if not props_to_extract:
        return []
    return [p["name"] for p in props_to_extract]


def validate_against_dictionary(gitops, data_dictionary):
    """
    Validates gitops.json configuration against a data dictionary
    Args:
        gitops (dict): gitops.json config
        data_dictionary (str): url of data dictionary

    Returns:
        ok(bool): whether the validation succeeded.

    """

    _, model = init_dictionary(data_dictionary)
    schema = model.dictionary.schema

    ok = True
    graphql = gitops["graphql"]
    for item in graphql["boardCounts"]:
        node_count = item["graphql"]
        # assumes form _{node}_count
        idx = node_count.rfind("_")
        node = node_count[1:idx]

        ok = ok and assert_and_log(
            schema.get(node) is not None,
            "Node: {} in graphql.boardCounts not found in dictionary".format(node),
        )

    for item in graphql["chartCounts"]:
        node_count = item["graphql"]
        # assumes form _{node}_count
        idx = node_count.rfind("_")
        node = node_count[1:idx]
        ok = ok and assert_and_log(
            schema.get(node) is not None,
            "Node: {} in graphql.chartCounts not found in dictionary".format(node),
        )

    for item in graphql.get("homepageChartNodes", []):
        node = item["node"]
        ok = ok and assert_and_log(
            schema.get(node) is not None,
            "Node: {} in graphql.homepageChartNodes not found in dictionary".format(
                node
            ),
        )

    return ok



import asyncio
import importlib
import re
import struct
import sys
from json import JSONDecoder, JSONDecodeError

NOT_WHITESPACE = re.compile(r"[^\s]")


async def stream_json(
    stdin, buf_size=65536, max_json=65536 * 16, decoder=JSONDecoder()
):
    buf = ""
    ex = None
    pending = True
    while pending:
        chunk = await stdin.read(buf_size)
        if chunk:
            buf += chunk.decode()
        else:
            pending = False
        pos = 0
        buf_len = len(buf)
        while pos < buf_len:
            match = NOT_WHITESPACE.search(buf, pos)
            if not match:
                break
            pos = match.start()
            if buf[pos] == "\x00":
                return
            start = pos
            try:
                obj, pos = decoder.raw_decode(buf, pos)
            except JSONDecodeError as e:
                if (not pending or len(buf) > max_json - buf_size) and (
                    ex is None or ex.pos == e.pos
                ):
                    pos = e.pos + 1
                    ex = None
                else:
                    print(
                        f"Cannot get JSON from chunk: {e}. Chunk:\n{buf}",
                        file=sys.stderr,
                    )
                    ex = e
                    break
            else:
                ex = None
                yield obj, buf[start:pos]
        buf = buf[pos:]
        if ex is not None:
            raise ex
    raise EOFError()


async def worker(loop, handle_row):
    stdin = asyncio.StreamReader()
    await loop.connect_read_pipe(lambda: asyncio.StreamReaderProtocol(stdin), sys.stdin)
    try:
        while True:
            lines = 0
            size = 0
            async for row, line in stream_json(stdin):
                lines += 1
                size += len(line)
                output = handle_row(row, line)
                if output:
                    print(output, file=sys.stderr)
            sys.stdout.buffer.write(struct.pack("QQ", lines, size))
            sys.stdout.flush()
    except EOFError:
        pass


def worker_main():
    script_module = sys.argv[1]
    handle_row = getattr(importlib.import_module(script_module), "handle_row")
    loop = asyncio.get_event_loop()
    loop.run_until_complete(worker(loop, handle_row))


if __name__ == "__main__":
    worker_main()



from dictionaryutils import DataDictionary, dictionary


def init_dictionary(url):
    d = DataDictionary(url=url)
    dictionary.init(d)
    # the gdcdatamodel expects dictionary initiated on load, so this can't be
    # imported on module level
    from gdcdatamodel import models as md

    return d, md



import os
import requests

from cdislogging import get_logger

logger = get_logger("Submit comments to PR", log_level="info")


def comment_on_pr(repository, pull_request_number, message, comments):
    token = os.environ["GITHUB_TOKEN"]
    headers = {"Authorization": "token {}".format(token)}

    repository = repository.strip("/")
    base_url = "https://api.github.com/repos/{}".format(repository)
    logger.info("Checking pull request: {} #{}".format(repository, pull_request_number))
    pr_comments_url = "{}/issues/{}/comments".format(base_url, pull_request_number)
    contents = ""
    for comment in comments:
        contents += "- {}\n".format(comment)
    full_comment = "# {}\n{}".format(message, contents)

    submit_comment(full_comment, headers, pr_comments_url)


def submit_comment(contents, headers, pr_comments_url):
    res = requests.post(pr_comments_url, json={"body": contents}, headers=headers)
    if res.status_code != 201:
        logger.error("Failed to write comment:", res.status_code, res.json())



import click
import json
import multiprocessing
import os
import yaml

from cdislogging import get_logger

from gen3utils.deployment_changes.generate_comment import (
    comment_deployment_changes_on_pr,
)
from gen3utils.manifest.manifest_validator import validate_manifest as val_manifest
from gen3utils.etl.etl_validator import validate_mapping
from gen3utils.gitops.gitops_validator import val_gitops
from gen3utils.utils import comment_on_pr


logger = get_logger("gen3utils", log_level="info")


CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))


@click.group()
def main():
    """Utils for Gen3 cdis-manifest management."""


@main.command()
@click.argument("etl_mapping_file", type=str, nargs=1, required=True)
@click.argument("manifest_file", type=str, nargs=1, required=True)
@click.argument("portal_config_file", type=str, nargs=1, required=True)
@click.argument("repository", type=str, nargs=1, required=False)
@click.argument("pull_request_number", type=int, nargs=1, required=False)
def validate_portal_config(
    etl_mapping_file, manifest_file, portal_config_file, repository, pull_request_number
):
    """Validate a PORTAL_CONFIG_FILE against the dictionary specified in the MANIFEST_FILE
    and an ETL_MAPPING_FILE"""

    logger.info("Validating portal config file {}".format(portal_config_file))
    dictionary_url = None
    with open(manifest_file, "r") as f1:
        manifest = json.loads(f1.read())
        dictionary_url = manifest.get("global", {}).get("dictionary_url")
        hostname = manifest.get("global", {}).get("hostname")
        if dictionary_url is None:
            logger.error("No dictionary URL in manifest {}".format(manifest_file))
            return

    recorded_errors, ok = val_gitops(
        dictionary_url, etl_mapping_file, portal_config_file
    )

    if recorded_errors:
        for err in recorded_errors:
            logger.error("  - {}".format(err))
        if repository and pull_request_number:
            if not "GITHUB_TOKEN" in os.environ:
                logger.warning("Missing GITHUB_TOKEN: not commenting on pull request.")
            else:
                message_header = (
                    "{}\n :x: etlMapping.yaml - gitops.json mismatch".format(hostname)
                )
                comment_on_pr(
                    repository,
                    pull_request_number,
                    message_header,
                    recorded_errors,
                )

    if not ok:
        raise AssertionError(
            "Portal configuration mapping validation failed. See errors in previous logs."
        )
    else:
        logger.info("  OK!")


@main.command()
@click.argument("manifest_files", type=str, nargs=-1, required=True)
def validate_manifest(manifest_files):
    """Validate one or more MANIFEST_FILES against a REQUIREMENTS_FILE."""

    requirements_file = os.path.join(CURRENT_DIR, "manifest", "validation_config.yaml")
    with open(requirements_file, "r") as f:
        requirements = yaml.safe_load(f.read())

    failed_validation = False
    for f_name in manifest_files:
        logger.info("Validating manifest {}".format(f_name))
        try:
            with open(f_name, "r") as f:
                cdis_manifest = json.loads(f.read())
            val_manifest(cdis_manifest, requirements)
        except AssertionError as e:
            logger.error("{}: {}".format(type(e).__name__, e))
            failed_validation = True
    if failed_validation:
        raise AssertionError("manifest validation failed. See errors in previous logs.")


@main.command()
@click.argument("etl_mapping_file", type=str, nargs=1, required=True)
@click.argument("manifest_file", type=str, nargs=1, required=True)
def validate_etl_mapping(etl_mapping_file, manifest_file):
    """Validate an ETL_MAPPING_FILE against the dictionary specified in the MANIFEST_FILE."""

    logger.info("Validating ETL mapping {}".format(etl_mapping_file))
    with open(manifest_file, "r") as f1:
        manifest = json.loads(f1.read())
        dictionary_url = manifest.get("global", {}).get("dictionary_url")
        if dictionary_url is None:
            logger.error("No dictionary URL in manifest {}".format(manifest_file))
            return

        logger.info("  Using dictionary: {}".format(dictionary_url))
        recorded_errors = validate_mapping(dictionary_url, etl_mapping_file, manifest)

        if recorded_errors:
            logger.error("  ETL mapping validation failed:")
            for err in recorded_errors:
                logger.error("  - {}".format(err))
            raise AssertionError(
                "ETL mapping validation failed. See errors in previous logs."
            )
        else:
            logger.info("  OK!")


@main.command()
@click.argument("repository", type=str, nargs=1, required=True)
@click.argument("pull_request_number", type=int, nargs=1, required=True)
def post_deployment_changes(repository, pull_request_number):
    """
    Comment on a pull request with any deployment changes when updating manifest services. Also comment a warning if a service is on a branch.
    """
    comment_deployment_changes_on_pr(repository, pull_request_number)


@main.command()
@click.argument("bucket")
@click.argument("prefix")
@click.argument("script")
@click.option("--region", default="us-east-1", show_default=True)
@click.option(
    "--access-key-id", default=os.environ.get("ACCESS_KEY_ID"), show_default=True
)
@click.option(
    "--secret-access-key",
    default=os.environ.get("SECRET_ACCESS_KEY"),
    show_default=True,
)
@click.option(
    "-c",
    "--concurrency",
    type=int,
    default=multiprocessing.cpu_count() + 1,
    show_default=True,
)
@click.option("--progress/--no-progress", default=True, show_default=True)
def s3log(*args, **kwargs):
    """Run SCRIPT in Gen3 logs under S3 BUCKET:PREFIX.

    The SCRIPT should be importable defining a method like this:

    \b
        def handle_row(obj, line):
            if 1 + 1 == 2:
                return line

    The returning results will be joined with newline into the stdout.
    """
    try:
        from gen3utils.s3log.s3log import S3Log
    except ImportError as e:
        print(e, '\nInstall with `poetry install --extras "s3log"` to run this command')
        exit(1)

    S3Log(*args, **kwargs).run()


if __name__ == "__main__":
    main()



import json
import os
from packaging import version
import re
import requests
import sys

from cdislogging import get_logger
import gen3git

from gen3utils.manifest.manifest_validator import version_is_branch
from gen3utils.utils import submit_comment

logger = get_logger("comment-deployment-changes", log_level="info")


# whitelist of services to ignore when checking tags and services on branch
# TODO: We're adding access-backend here just to unblock the gen3utils check in gitops-qa's .travis.yml
# This repo is private and there seems to be an issue with gen3utils where it is not picking up the GITHUB_TOKEN
IGNORED_SERVICES = [
    "ambassador",
    "aws-es-proxy",
    "fluentd",
    "jupyterhub",
    "nb2",
    "access-backend",
]

# update this config if the service name in the manifest "versions"
# block is not the same as the repo name. services that are not
# listed here are assumed to be in repo uc-cdis/<service name>.
SERVICE_TO_REPO = {
    "awshelper": "cloud-automation",
    "dashboard": "gen3-statics",
    "portal": "data-portal",
    "revproxy": "docker-nginx",
    "spark": "gen3-spark",
    "wts": "workspace-token-service",
    "metadata": "metadata-service",
    "covid19-etl": "covid19-tools",
    "covid19-notebook-etl": "covid19-tools",
    "covid19-bayes": "covid19-tools",
    "data-ingestion-pipeline": "bdcat-data-ingestion",
    "datareplicate": "dcf-dataservice",
}


def comment_deployment_changes_on_pr(repository, pull_request_number):
    """
    Gets the deployments changes for the specified pull request and write them in a comment, along with a warning if any service is on a branch.

    Args:
        repository (str): "<user>/<repo>"
        pull_request_number (str)
    """
    if "GITHUB_TOKEN" in os.environ:
        token = os.environ["GITHUB_TOKEN"]
        headers = {"Authorization": "token {}".format(token)}
    else:
        logger.warning(
            "Missing GITHUB_TOKEN: NOT commenting deployment changes on pull request. Continuing; but might fail later if the PR is not accessible."
        )
        token = None
        headers = {}

    repository = repository.strip("/")
    base_url = "https://api.github.com/repos/{}".format(repository)
    logger.info("Checking pull request: {} #{}".format(repository, pull_request_number))
    pr_files_url = "{}/pulls/{}/files".format(base_url, pull_request_number)
    pr_comments_url = "{}/issues/{}/comments".format(base_url, pull_request_number)

    # get list of files from PR
    files = requests.get(pr_files_url, headers=headers).json()
    if not isinstance(files, list):
        logger.error(files)
        raise Exception("Unable to get PR files")

    # only keep manifest.json files
    manifest_files = [f for f in files if f["filename"].endswith("manifest.json")]
    if not manifest_files:
        logger.info("No manifest files to check - exiting")
        return

    # for each modified manifest file, compare versions to master versions
    logger.info("Checking manifest files")
    full_comment = ""
    for file_info in manifest_files:
        logger.info("- {}".format(file_info["filename"]))
        url = file_info["raw_url"]
        old_file, new_file = get_files(get_master_url(url), url, headers)
        new_versions_block = new_file.get("versions", {})
        new_is_nde_portal = "data-ecosystem-portal" in new_versions_block.get(
            "portal", ""
        )

        for service_name in IGNORED_SERVICES:
            new_versions_block.pop(service_name, None)

        if old_file:
            old_versions_block = old_file.get("versions", {})
            for service_name in IGNORED_SERVICES:
                old_versions_block.pop(service_name, None)

            old_is_nde_portal = "data-ecosystem-portal" in old_versions_block.get(
                "portal", ""
            )
            # we only compare portal versions from the same repo
            compared_versions = compare_versions_blocks(
                old_versions_block,
                new_versions_block,
                old_is_nde_portal == new_is_nde_portal,
            )
            deployment_changes = get_deployment_changes(
                compared_versions, token, new_is_nde_portal
            )
        else:
            # if this is a new file, there's nothing to compare
            deployment_changes = {}
        contents = generate_comment(
            deployment_changes, check_services_on_branch(new_versions_block)
        )
        if contents:
            full_comment += "# {}\n{}".format(file_info["filename"], contents)
    if full_comment:
        logger.info(full_comment)
        submit_comment(full_comment, headers, pr_comments_url)


def get_master_url(modified_file_url):
    """
    Returns the URL to the current version of the modified file, assuming
    the current version is on branch "master".

    Args:
        modified_file_url (str): URL to the modified version of the file
    """
    parts = modified_file_url.split("/")
    hash_index = parts.index("raw") + 1
    parts[hash_index] = "master"
    return "/".join(parts)


def get_files(master_url, pr_url, headers):
    """
    Returns the contents of the current version of the file and
    of the modified version of the file in a tuple.

    Args:
        master_url (str): URL to the current version of the file
        pr_url (str): URL to the modified version of the file
        headers (dict): Authorization heading with a token with read and
            write access to the repo
    """
    old_res = requests.get(master_url, headers=headers)
    new_res = requests.get(pr_url, headers=headers)
    if (
        old_res.status_code != 200 and old_res.status_code != 404
    ) or new_res.status_code != 200:
        raise Exception(
            f"Unable to get files:\n{master_url} {old_res.status_code}\n{pr_url} {new_res.status_code}"
        )

    old_file = None if old_res.status_code == 404 else old_res.json()
    new_file = new_res.json()
    return old_file, new_file


def compare_versions_blocks(old_versions_block, new_versions_block, check_portal):
    """
    Returns a dict:
    {
        <service name>: { "old": <version>, "new": <version> }
    }
    """

    services = list(
        set(old_versions_block.keys()).union(set(new_versions_block.keys()))
    )
    services.sort()

    res = {}
    for service in services:
        if service == "portal" and not check_portal:
            # skip if portal versions are not from the same repo
            continue

        old_version = old_versions_block.get(service)
        new_version = new_versions_block.get(service)
        if (
            not old_version
            or not new_version
            or (
                "quay.io/cdis" not in old_version
                and "dkr.ecr.us-east-1.amazonaws.com" not in old_version
            )
            or (
                "quay.io/cdis" not in new_version
                and "dkr.ecr.us-east-1.amazonaws.com" not in new_version
            )
            or len(old_version.split(":")) < 2
            or len(new_version.split(":")) < 2
        ):
            # new service, or deleted service: nothing to compare.
            # non-CTDS repo: no deployment changes to get.
            # version without ":" is not usable.
            continue
        old_version = old_version.split(":")[1]
        new_version = new_version.split(":")[1]
        if old_version != new_version:
            res[service] = {"old": old_version, "new": new_version}

    logger.info("Updates: {}".format(json.dumps(res, indent=2)))
    return res


def get_deployment_changes(versions_dict, token, is_nde_portal):
    """
    Uses the gen3git utility to get the release notes between the old and new
    versions for each service, and returns the deployment changes only.

    Args:
        versions_dict (dict):
            {
                <service name>: { "old": <version>, "new": <version> }
            }
        token (string): token with read and write access to the repo
    """

    class Gen3GitArgs(object):
        def __init__(self, repo, from_tag, to_tag):
            self.github_access_token = token
            self.repo = repo
            self.from_tag = from_tag
            self.to_tag = to_tag

    res = {}
    for service, versions in versions_dict.items():
        # only get the deployment changes if the new version is more
        # recent than the old version. ignore services on a branch
        if not version_is_branch(
            versions["old"], release_tag_are_branches=False
        ) and not version_is_branch(versions["new"], release_tag_are_branches=False):
            # by default, assume the code lives in repo uc-cdis/<service name>
            repo_name = SERVICE_TO_REPO.get(service, service)

            # repo names special cases
            if service == "portal" and is_nde_portal:
                repo_name = "data-ecosystem-portal"

            repo_name = "uc-cdis/" + repo_name
            args = Gen3GitArgs(repo_name, versions["old"], versions["new"])
            try:
                release_notes = gen3git.main(args)
                if not release_notes:
                    raise Exception("gen3git did not return release notes")
            except:
                logger.error(
                    "While checking service '{}', repo '{}', unable to get release notes with gen3git:".format(
                        service, repo_name
                    )
                )
                raise
            notes = release_notes.get("deployment changes")
            if notes:
                res[service] = update_pr_links(repo_name, notes)
    return res


def check_services_on_branch(versions_block):
    """
    Returns the list of all services that are on a branch.
    """
    services_on_branch = []
    for service in versions_block:
        version = versions_block.get(service)
        if "quay.io/cdis" not in version or len(version.split(":")) < 2:
            # ignore non-CTDS repos.
            # version without ":" is not usable.
            continue
        version = version.split(":")[1]
        if version_is_branch(version, release_tag_are_branches=False):
            services_on_branch.append(service)
    return services_on_branch


def update_pr_links(repo_name, notes_list):
    """
    Replace the internal repo PR link with the external repo PR link
    in each release note.
    """
    result = []
    matcher = re.compile(".*\(#(?P<pr_number>[0-9]+)\)$")
    # e.g. gets the PR number ("12") from "some description (#12)"
    for note in notes_list:
        match = matcher.match(note)
        if match:
            internal_pr_number = "#{}".format(match.groupdict()["pr_number"])
            external_pr_number = "{}{}".format(repo_name, internal_pr_number)
            result.append(note.replace(internal_pr_number, external_pr_number))
        else:
            result.append(note)
    return result


def generate_comment(deployment_changes, services_on_branch):
    # TODO: edit the previous comment instead of posting a new one
    contents = ""
    if services_on_branch:
        contents += "## :warning: Services on branch\n- {}\n".format(
            "\n- ".join(services_on_branch)
        )
    if deployment_changes:
        contents += "## Deployment changes\n"
        for service, items in deployment_changes.items():
            contents += "- {}\n  - {}\n".format(service, "\n  - ".join(items))
    return contents



class MappingError(object):
    def __init__(self, message, type):
        self.message = "{} error: {}".format(type, message)
        self.type = type

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "{}".format(self.message)

    def __eq__(self, other):
        return (
            isinstance(other, MappingError)
            and self.type == other.type
            and self.message == other.message
        )


class PropertiesError(MappingError):
    def __init__(self, message):
        super(PropertiesError, self).__init__(message, "Properties")


class PathError(MappingError):
    """
    Path error specifies the path that does not exist in checking dicitonary
    :rtype: PathError object
    """

    def __init__(self, path):
        super(PathError, self).__init__(
            "{} does not exist in this dictionary".format(path), "Path"
        )


class FieldError(MappingError):
    def __init__(self, message):
        super(FieldError, self).__init__(message, "Field")


class FieldSyntaxError(MappingError):
    def __init__(self, field):
        message = "Required field [{}] not found".format(field)
        super(FieldSyntaxError, self).__init__(message, "FieldSyntaxError")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gen3utils',
 'gen3utils.deployment_changes',
 'gen3utils.etl',
 'gen3utils.gitops',
 'gen3utils.manifest',
 'gen3utils.s3log']

package_data = \
{'': ['*']}

install_requires = \
['cdislogging>=1.0.0,<1.1.0',
 'click',
 'dictionaryutils>=3.0.0,<3.1.0',
 'gen3datamodel>=3.0.0,<3.1.0',
 'gen3dictionary>=2.0.1,<2.1.0',
 'gen3git>=0.3.10,<0.4.0',
 'packaging>=20.0,<20.1',
 'psqlgraph>=3.0,<3.1',
 'pyyaml>=5.4,<5.5',
 'six>=1.12.0,<1.13.0']

extras_require = \
{':extra == "s3log"': ['aiobotocore==0.11.1']}

entry_points = \
{'console_scripts': ['gen3utils = gen3utils.main:main']}

setup_kwargs = {
    'name': 'gen3utils',
    'version': '0.6.13',
    'description': 'Gen3 Library Template',
    'long_description': '# gen3utils\n\nUtils for Gen3 commons management\n\n## Install with pip\npip install gen3utils\n\n## manifest.json validation\n\nValidate one or more `manifest.json` files:\n```\ngen3utils validate-manifest cdis-manifest/*/manifest.json\n```\n\nThe validation settings can be updated by modifying [this file](gen3utils/manifest/validation_config.yaml).\n\n## etlMapping.yaml validation\n\nValidate an `etlMapping.yaml` file against the dictionary URL specified in a `manifest.json` file:\n```\ngen3utils validate-etl-mapping etlMapping.yaml manifest.json\n```\n\n## Portal Configuration (gitops.json) validation\n\nValidate a `gitops.json` file against the dictionary URL specified in a `manifest.json` file and an etlMapping.yaml file. Adds a comment to a pull request listing all the errors encountered when validating against etlMapping.yaml\n```\ngen3utils validate-portal-config etlMapping.yaml manifest.json gitops.json <username>/<repository> <pull request number>\n```\nTo run without making a pull request comment\n```\ngen3utils validate-portal-config etlMapping.yaml manifest.json gitops.json\n```\n\n\n## Comment on a PR with any deployment changes when updating manifest services\n\nThe command requires the name of the repository, the pull request number and **a `GITHUB_TOKEN` environment variable** containing a token with read and write access to the repository. It also comments a warning if a service is pinned on a branch.\n```\npip install gen3utils\ngen3utils post-deployment-changes <username>/<repository> <pull request number>\n```\n\n## Log parser for CTDS log pipeline\n\n```\npip install gen3utils\ngen3utils s3log --help\ngen3utils s3log [OPTIONS] BUCKET PREFIX SCRIPT\n```\n\nRun `SCRIPT` in Gen3 logs under S3 `BUCKET:PREFIX`. The `SCRIPT` should be importable defining a method like this:\n```\ndef handle_row(obj, line):\n    if 1 + 1 == 2:\n        return line\n```\n\nFor example, to process logs in bucket `my-commons-logs` at prefix `my-logs` with a `gen3utils/script.py` file:\n```\npip install gen3utils\ngen3utils s3log my-commons-logs my-logs gen3utils.script\n```\n\n## Running tests locally\n\n```\npoetry install -vv\npoetry run pytest -vv ./tests\n```\n',
    'author': 'CTDS UChicago',
    'author_email': 'cdis@uchicago.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/uc-cdis/gen3utils',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from packaging import version
import re
from cdislogging import get_logger
from gen3utils.assertion import assert_and_log

logger = get_logger("validate-manifest", log_level="info")


def validate_manifest(manifest, validation_requirement):
    """
    Runs all the validation checks against a manifest.json file.

    Args:
        manifest (dict): Contents of manifest.json file.
        validation_requirement (dict): Contents of validation_config.yaml file.
    """

    # remove services in avoid in validation_config which don't need validation
    hostname = manifest["global"].get("hostname")
    if hostname in validation_requirement.get("avoid", []):
        services_to_skip = validation_requirement["avoid"][hostname]
        for s in services_to_skip:
            del manifest["versions"][s]

    ok = True
    if "block" in validation_requirement:
        ok = validate_manifest_block(manifest, validation_requirement["block"]) and ok
    if "versions" in validation_requirement:
        ok = (
            versions_validation(
                manifest.get("versions"), validation_requirement["versions"]
            )
            and ok
        )
    ok = misc_validations(manifest) and ok

    if not ok:
        raise AssertionError(
            "cdis-manifest validation failed. See errors in previous logs."
        )
    logger.info("OK")


def validate_manifest_block(manifest, blocks_requirements):
    """
    Validates blocks in cdis-manifest.

    Args:
        manifest (dict): Contents of manifest.json file.
        blocks_requirements (dict): the "block" requirement under
            validation_config.yaml. The keys of the dict are the service names
            and the values are the requirements for the block associated to
            the service.

    Return:
        ok(bool): whether the validation succeeded.
    """
    ok = True

    for service_name, block in blocks_requirements.items():
        if service_name in manifest["versions"]:
            # Validation for all services has requirement in validation_config.
            block_requirement_version = get_manifest_version(
                manifest["versions"], service_name
            )
            is_branch = version_is_branch(block_requirement_version)
            should_check_has = type(block) == dict and "has" in block and not is_branch
            if type(block) == dict and "version" in block and not is_branch:
                # If we have version requirement for a service, min or max is required.
                # We are not validating branch or master branch
                # min: Version in manifest is equal or greater than the verson in validation_config
                # max: Version in manifest is smaller than the verson in validation_config

                min_version = block["version"].get("min")
                max_version = block["version"].get("max")
                if min_version:
                    min_version = version.parse(min_version)
                if max_version:
                    max_version = version.parse(max_version)

                if min_version and max_version:
                    should_check_has = (
                        min_version <= block_requirement_version < max_version
                    )
                elif min_version:
                    should_check_has = block_requirement_version >= min_version
                elif max_version:
                    should_check_has = block_requirement_version < max_version

            if should_check_has:
                # Validation to check if a service has a specific key in its block in cdis-manfiest
                error_msg = "{} is missing in {} block or {} block is missing".format(
                    block["has"], service_name, service_name
                )

                ok = (
                    assert_and_log(
                        service_name in manifest
                        and block["has"] in manifest[service_name]
                        or service_name not in manifest
                        and block.get("optional"),
                        error_msg,
                    )
                    and ok
                )

            if block is True:
                # Validation to check if a block exists in cdis-manifest
                ok = (
                    assert_and_log(
                        service_name in manifest,
                        service_name + " block is missing in cdis-manifest",
                    )
                    and ok
                )
    return ok


def get_manifest_version(
    manifest_versions, service, release_tag_are_branches=True, warn=True
):
    """
    Get the service version from cdis-manifest
        Arg:
            service: microservice name
            manifest_versions: the versions block from manifest.json
        Return:
            parsed microservice version (or None if service not in manifest)
            (or version string if unable to parse or version is a branch)
    """
    for manifest_version in manifest_versions:
        if manifest_version == service:
            # if this fails, make sure you check the service is in the manifest before
            # calling get_manifest_version()! it should never happen
            service_line = manifest_versions[service]
            service_version = service_line.split(":")[1]
            if version_is_branch(service_version, release_tag_are_branches):
                if warn:
                    logger.warning(
                        "{} is on a branch ({}): not validating, returning string type".format(
                            service, service_version
                        )
                    )
                return service_version
            try:
                return version.parse(service_version)
            except:
                if warn:
                    logger.warning(
                        "Cannot parse version '{}', returning string type".format(
                            service_version
                        )
                    )
                return service_version


def version_is_branch(version, release_tag_are_branches=True):
    """
    Args:
        version (string)
        release_tag_are_branches (bool): whether release tags in format
            <dddd.dd> should be considered branches or not.
            - For manifest validation, we want to skip validation for release
            tags because the semantic versions comparison would not work.
            - For checking deployment changes, we do want release tags to be
            included like other tags.

    Returns:
        bool: True if version only contains digits and dots BUT is
            not a release tag in format <dddd.dd>
    """
    # check if it looks like semantic versioning
    reg = re.compile("^[0-9]+[.[0-9]+]*$")
    is_branch = not bool(reg.match(str(version)))

    # check if it's a release tag
    if not is_branch and release_tag_are_branches:
        reg = re.compile("^[0-9]{4}.[0-9]{2}$")
        is_branch = bool(reg.match(str(version)))

    return is_branch


def versions_validation(manifest_versions, versions_requirements):
    """
    Validates versions in cdis-manifest

    Arg:
        manifest_versions: manifest.json "versions" section
        versions_requirements: the "versions" requirement under validation_config.yaml

    Return:
        ok(bool): whether the validation succeeded.
    """
    ok = True

    for versions_requirement in versions_requirements:

        requirement_list = versions_requirement["needs"]
        requirement_key_list = list(requirement_list.keys())
        requirement_key = list(versions_requirement)[0]
        actual_version = get_manifest_version(manifest_versions, requirement_key)

        if requirement_key in manifest_versions and not version_is_branch(
            actual_version
        ):

            required_version = versions_requirement[requirement_key]

            # If the first service set to * under validation_config versions, other services should be in the manifest
            # The second condition is ignoring branch on sevice. WHICH IS NOT GOO. Added a warning in the log
            if required_version == "*":
                for required_service in requirement_key_list:
                    ok = (
                        assert_and_log(
                            required_service in manifest_versions,
                            required_service + " is missing in manifest.json",
                        )
                        and ok
                    )

            elif (
                "min" not in required_version
                and "max" not in required_version
                and version.parse(required_version) <= actual_version
            ):
                # If the first service set to a specific version in validation_config, other services should matches the version requirements
                ok = (
                    version_requirement_validation(
                        requirement_list,
                        requirement_key_list,
                        manifest_versions,
                        "{} {}".format(requirement_key, actual_version),
                    )
                    and ok
                )
            elif (
                "min" in required_version
                and "max" in required_version
                and version.parse(required_version["min"])
                <= actual_version
                < version.parse(required_version["max"])
            ):
                # if service is min_requirement <= service_version < max_requirement, other services should matches the version requirements
                ok = (
                    version_requirement_validation(
                        requirement_list,
                        requirement_key_list,
                        manifest_versions,
                        "{} {}".format(requirement_key, actual_version),
                    )
                    and ok
                )

    return ok


def version_requirement_validation(
    service_requirement, requirement_key_list, manifest_versions, current_validation
):
    """
    Validates version matches the requirement for a specific service

    Args:
    service_requirement: service name and its version requirement
    requirement_key_list: list of service name which need validation
    manifest_versions: versions block from manifest
    current_validation (str): service name and version that are currently being validated

    Return:
        ok(bool): whether the validation succeeded.
    """
    ok = True

    for required_service in requirement_key_list:
        actual_version = get_manifest_version(manifest_versions, required_service)
        if not actual_version:
            logger.error(
                'Service "{}" not in manifest but required to validate "{}" with "{}"'.format(
                    required_service, current_validation, service_requirement
                )
            )
            ok = False
            continue

        if version_is_branch(actual_version):
            # ignoring service on branch - user was already informed of this
            # by the log in `manifest_version()`
            continue

        if (
            "min" not in service_requirement[required_service]
            and "max" not in service_requirement[required_service]
        ):
            ok = (
                assert_and_log(
                    version.parse(service_requirement[required_service])
                    <= actual_version,
                    'Service "{}" version "{}" does not respect requirement "{}" for "{}"'.format(
                        required_service,
                        actual_version,
                        service_requirement,
                        current_validation,
                    ),
                )
                and ok
            )
        elif (
            "min" in service_requirement[required_service]
            and "max" in service_requirement[required_service]
        ):
            ok = (
                assert_and_log(
                    version.parse(service_requirement[required_service]["min"])
                    <= actual_version
                    < version.parse(service_requirement[required_service]["max"]),
                    'Service "{}" version "{}" does not respect requirement "{}" for "{}"'.format(
                        required_service,
                        actual_version,
                        service_requirement,
                        current_validation,
                    ),
                )
                and ok
            )
    return ok


def misc_validations(manifest):
    """
    Handles all the ad-hoc validations which are not structurally grouped in the validation_config.yml

    Args:
    manifest: manifest.json file as a dictionary

    Return:
        ok(bool): whether the validation succeeded.
    """
    ok = True

    # Hatchery Service needs network policies to be turned on in the global block
    if "hatchery" in manifest["versions"]:
        shouldSkip = manifest["global"]["hostname"] in [
            # the NDE has a special setup; network policies must be turned off
            # to avoid cross-namespace communication issues
            "qa-nde.planx-pla.net",
            "niaiddata.org",
        ]
        if shouldSkip is False and (
            "netpolicy" not in manifest["global"]
            or manifest["global"]["netpolicy"] != "on"
        ):
            logger.error(
                "Hatchery needs netpolicy==on in the global block of the manifest"
            )
            ok = False
    return ok



from cdislogging import get_logger

logger = get_logger("gen3utils", log_level="info")


def assert_and_log(assertion_success, error_message):
    """
    If an assertion fails, logs the provided error message and updates
    the global variable "failed_validation" for future use.

    Args:
        assertion_success (bool): result of an assertion.
        error_message (str): message to display if the assertion failed.

    Return:
        assertion_success(bool): result of the assertion.
    """
    if not assertion_success:
        logger.error(error_message)
    return bool(assertion_success)
